# $Id: database.py 492 2007-11-13 18:39:30Z jcea $


VERSION_DB="2007111302"

import globales

# Chanchullo para los "tests"
do_backup_users=True

timezones={}

def normaliza_nick(nick) :
  return nick.lower()

def init_database(monitor) :
  @monitor
  def init(conn) :
    root=conn.get_root()
    version=root.get("version del foro",None)
    if version==VERSION_DB : return
    if version :
      import upgrade
      upgrade.do(conn)
      version=root.get("version del foro",None)
      if version==VERSION_DB : return
      if version :
        raise RuntimeError("Version incompatible de la base de datos: Queremos '%s' y tenemos '%s'" %(VERSION_DB,version))

    root["version del foro"]=VERSION_DB

    from durus.btree import BTree
    from durus.persistent_dict import PersistentDict
    from durus.persistent_list import PersistentList
    from durus.persistent_set import PersistentSet

    usuarios=PersistentDict()
    usuarios["usuarios"]=BTree()
    usuarios["num_usuarios"]=0
    usuarios["pending_sign_ins"]=BTree()
    root["usuarios"]=usuarios

    metatags=PersistentDict()
    metatags["publicos"]=PersistentSet()
    metatags["categorias"]=PersistentList()
    metatags["metatags"]=BTree()
    metatags["num_metatags"]=0
    root["metatags"]=metatags

    grupos=PersistentDict()
    grupos["grupos"]=PersistentDict()
    grupos["num_grupos"]=0
    root["grupos"]=grupos

    hilos=PersistentDict()
    hilos["hilos"]=BTree()
    hilos["num_hilos"]=0
    root["hilos"]=hilos

    mensajes=PersistentDict()
    mensajes["mensajes"]=BTree()
    mensajes["num_mensajes"]=0
    root["mensajes"]=mensajes

    root["outgoing emails"]=BTree()

    seguimiento_no_leidos=PersistentDict()
    seguimiento_no_leidos["hilo2last_msg"]=BTree()
    seguimiento_no_leidos["last_msg2hilo"]=BTree()
    root["seguimiento_no_leidos"]=seguimiento_no_leidos

    root["punto_de_lectura_global"]=BTree()

    root["static"]=PersistentDict()
    root["static"]["avatars"]=BTree()
    root["static"]["smileys"]=BTree()
    root["static"]["css"]=BTree()
    root["static"]["js"]=BTree()
    root["static"]["icons"]=BTree()
    root["smileys"]=PersistentList()

    conf=PersistentDict()
    conf["thread_len"]=10
    conf["allow_anonymous"]=True
    conf["allow_private_messages"]=False
    conf["http_max_clients"]=16
    conf["http_initial_timeout"]=30
    conf["http_keep_alive_timeout"]=5
    conf["user_sign_in_mode"]=0  # Only "master" can create new users
    conf["openid_support"]=False # Because You could not have the libs!
    conf["allow_bbcode"]=True
    conf["allow_html"]=True
    conf["allow_avatar_change"]=False # Because You could not have the libs!
    conf["base_url"]=""
    conf["smtp_server"]=""
    conf["smtp_server_port"]=25
    conf["smtp_auth_user"]=""
    conf["smtp_auth_passwd"]=""
    conf["smtp_use_tls"]=False
    conf["smtp_email_sender"]=""
    conf["admin_email_address"]=""
    conf["tz"]="Europe/Madrid"
    conf["google_verification"]=""
    root["config"]=conf

  @monitor
  def load_initial_config(conn) :
    conf=dict(conn.get_root()["config"]) # Force object load
    return load_new_config(conf)

  init()
  load_initial_config()


def load_new_config(conf) :
  import globales
  for k,v in conf.items() :
    setattr(globales,k,v)


def usuario_get_tz(conn,nick) :
  root=conn.get_root()
  usuario=root["usuarios"]["usuarios"].get(normaliza_nick(nick),None)
  if usuario :
    tz=usuario["tz"]
    if tz in timezones :
      tz=timezones[tz]
    else : # RACES have no bad consecuences
      import pytz
      tz=timezones[tz]=pytz.timezone(tz)
  else :
    tz=globales.tz

  return tz

def usuario_new_cookie(conn,nick) :
  root=conn.get_root()
  import random,sys
  cookie=str(random.randint(0,sys.maxint))+str(random.randint(0,sys.maxint))
  root["usuarios"]["usuarios"][normaliza_nick(nick)]["cookie"]=cookie
  return cookie

# Si suministramos una cookie, de ahi se saca el
# usuario y se comprueba su identidad.
# Si pasamos nick/clave, de ahi comprobamos tambien su identidad
# Si SOLO pasamos un nick, solo comprobamos si el usuario existe,
# pero no verificamos identidad.
def usuario_verifica(conn,nick=None,clave=None,cookie=None,force_login=False) :
  root=conn.get_root()

  if clave or force_login :
    usuario=root["usuarios"]["usuarios"].get(normaliza_nick(nick),None)
    if usuario and ((clave==usuario["clave"]) or force_login) :
      if "no_login" in usuario : return None
      cookie=usuario_new_cookie(conn,nick)
      nick=usuario["nick"]
      cookie="%s-%s" %(nick,cookie)
      return nick,cookie
  elif cookie :
    i=cookie.rfind("-")
    if i<=0 : return None # Al menos un caracter
    nick=cookie[:i]
    cookie=cookie[i+1:] # Nos saltamos el "-"
    usuario=root["usuarios"]["usuarios"].get(normaliza_nick(nick),None)
    if usuario and (cookie==usuario["cookie"]) :
      return usuario["nick"],None
  elif nick :
    return normaliza_nick(nick) in root["usuarios"]["usuarios"]

  return None

def set_usuario_no_login(conn,nick) :
  root=conn.get_root()
  usuario=root["usuarios"]["usuarios"][normaliza_nick(nick)]
  usuario["no_login"]=True

def unset_usuario_no_login(conn,nick) :
  root=conn.get_root()
  usuario=root["usuarios"]["usuarios"][normaliza_nick(nick)]
  if "no_login" in usuario :
    del usuario["no_login"]

def get_static(conn,section,name) :
  root=conn.get_root()
  section=root["static"].get(section)
  if not section : return None
  data=section.get(name)
  return data # If None -> not found

def get_static_metadata(conn,section,name) :
  metadata=get_static(conn,section,name)
  if metadata : metadata=metadata[0]
  return metadata

def save_static(conn,section,data,metadata,update=False) :
  def date_now() :
    import time
    weekdayname = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
    monthname = [None,
                 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']
    year, month, day, hh, mm, ss, wd, y, z = time.gmtime()
    ts = "%s, %02d %3s %4d %02d:%02d:%02d GMT" % (
            weekdayname[wd],
            day, monthname[month], year,
            hh, mm, ss)
    return ts

  from durus.persistent_list import PersistentList
  root=conn.get_root()
  import md5
  hash=md5.new(data).hexdigest()
  d=root["static"][section]
  if (hash not in d) or update : 
    metadata["last modified"]=date_now()
    d[hash]=(metadata,PersistentList([data]))
  return hash

def check_image(img) :
  metadata={}
  from PIL import Image
  from cStringIO import StringIO
  f=StringIO(img)
  try :
    im=Image.open(f)
    metadata["size"]=im.size
    format=im.format
    if format not in ("PNG","JPEG","GIF") :
      return ("Solo admitimos im&aacute;genes en formato PNG, JPEG o GIF",None)
    metadata["content-type"]="image/%s" %format.lower()
    im.verify()
  except :
    return ("El fichero parece corrupto o no se corresponde a una imagen",None)

  return (None,metadata)

def smileys_update(conn,smileys,directory) :
  import os,os.path
  root=conn.get_root()

  files={}
  for i in smileys :
    files[i[1]]=None # Eliminate duplicates

  for i in files.keys() :
    i2=i.replace("/",os.sep)
    img=open(os.path.join(directory,i2),"rb").read()  # Binary data *must* be read as binary under MS-Windows
    result,img_metadata=check_image(img)
    if result : raise RuntimeError("ERROR al cargar los iconos:",result)
    files[i]=save_static(conn,"smileys",img,img_metadata,update=False)

  from durus.persistent_list import PersistentList
  if len(set([i[0] for i in smileys]))!=len(smileys) : raise RuntimeError("ERROR al cargar los iconos: duplicados")
  smiley_list=PersistentList()
  smileys.sort(cmp=lambda x,y: cmp((len(y[0]),y),(len(x[0]),x))) # Longest prefixes first
  for i in smileys :
    smiley_list.append((i[0],files[i[1]]))

  root["smileys"]=smiley_list

# The list given is "longest prefixes first"
def get_smiley_list(conn) :
  root=conn.get_root()
  return list(root["smileys"])

def avatar_update(conn,nick,avatar) :
  if len(avatar)>125*1024 :
    return "El avatar mide m&acute;s de 125Kbytes"
  result,avatar_metadata=check_image(avatar)
  if result : return result

  x,y=avatar_metadata["size"]
  if x>150 or y>150 :
    return "Solo admitimos im&aacute;genes de un m&aacute;ximo de 150*150 pixels"

  name=save_static(conn,"avatars",avatar,avatar_metadata,update=False)
  assert name

  root=conn.get_root()
  usuarios=root["usuarios"]["usuarios"]

  usuario=usuarios.get(normaliza_nick(nick))
  assert usuario

  usuario["avatar"]=name

def usuario_update(conn,nick,clave=None,OpenIDs=None,TZ=None) :
  root=conn.get_root()
  usuarios=root["usuarios"]["usuarios"]

  usuario=usuarios.get(normaliza_nick(nick))
  assert usuario

  if clave :
    usuario["clave"]=clave

  if OpenIDs!=None :
    old_OpenIDs=usuario["OpenID"]
    if old_OpenIDs!=OpenIDs : # Solo grabamos si ha habido cambios
      old_OpenIDs.clear() # Este tipo es persistente, asi que lo reutilizamos
      old_OpenIDs.update(OpenIDs)

  if (TZ!=None) and (TZ!=usuario["tz"]) :
    usuario["tz"]=TZ

def usuario_ask_new_email(conn,nick,email) :
  nick_normalizado=normaliza_nick(nick)
  usuario=conn.get_root()["usuarios"]["usuarios"][nick_normalizado]
  import random,sys
  nonce=str(random.randint(0,sys.maxint))+str(random.randint(0,sys.maxint))
  usuario["new email"]=(nonce,email)
  return nonce

def usuario_confirm_new_email(conn,nick,nonce) :
  nick_normalizado=normaliza_nick(nick)
  usuario=conn.get_root()["usuarios"]["usuarios"][nick_normalizado]
  nonce_orig,email=usuario.get("new email",(None,None))

  if nonce_orig!=nonce : return False

  del usuario["new email"]
  usuario["email"]=email
  return True

def get_usuario_pending(conn,nonce) :
  root=conn.get_root()
  usuarios_pending=root["usuarios"]["pending_sign_ins"]
  v=usuarios_pending.get(nonce,None)
  if not v : return None
  del usuarios_pending[nonce]
  return v[2:]  # We don't return the time

# If we add new params, we must update "usuario_add".
# Also, we would need to update the confirmation code (via URL).
def usuario_add_pending(conn,nick,datos,clave=None,email="",OpenIDs=set()) :
  root=conn.get_root()
  usuarios=root["usuarios"]
  usuarios_pending=usuarios["pending_sign_ins"]

  nick_normalizado=normaliza_nick(nick)
  assert nick_normalizado not in usuarios["usuarios"]

  import time,random,sys
  while True :
    k=str(random.randint(0,sys.maxint))
    if k not in usuarios_pending : break

  usuarios_pending[k]=(time.time(),False,nick,datos,clave,email,OpenIDs)

  return k

# If we add new params, we must update "usuario_add_pending".
def usuario_add(conn,nick,datos,clave=None,email="",OpenIDs=set()) :
  from durus.btree import BTree
  from durus.persistent_dict import PersistentDict
  from durus.persistent_set import PersistentSet
  import time

  root=conn.get_root()
  usuarios=root["usuarios"]

  nick_normalizado=normaliza_nick(nick)
  assert nick_normalizado not in usuarios["usuarios"]

  usuarios["num_usuarios"]+=1

  if not clave :
    import random,sys
    clave=str(random.randint(0,sys.maxint))

  bt=BTree()
  for num_metatag in root["metatags"]["publicos"] :
    bt[num_metatag]=BTree()

  metatag_privado=metatag_add(conn,"Mensajes privados de "+nick,private=True)
  bt[metatag_privado]=BTree()

  usuarios["usuarios"][nick_normalizado]=PersistentDict(
                       {"ultimo mensaje conocido":root["mensajes"]["num_mensajes"],
                        "punto de lectura no leidos":bt,
                        "punto de lectura":BTree(),
                        "TS creacion":time.time(),
                        "TS ultima entrada":0,
                        "num mensajes":0,
                        "mensajes":BTree(),
                        "avatar":None,
                        "datos":datos,
                        "clave":clave,
                        "OpenID":PersistentSet(OpenIDs),
                        "email":email,
                        "cookie":None,
                        "monitor metatags":BTree(),
                        "monitor hilos":BTree(),
                        "metatag_privado":metatag_privado,
                        "tz":globales.tz,
                        "grupos":set(),
                        "nick":nick})

  toggle_monitor_metatag(conn,nick,metatag_privado)

  if do_backup_users :
    f=open("claves.backup","w")
    for i,j in conn.get_root()["usuarios"]["usuarios"].items() :
      print >>f,i,j["clave"]
    f.close()

def usuario_get(conn,nick) :
  root=conn.get_root()
  usuarios=root["usuarios"]

  nick_normalizado=normaliza_nick(nick)

  return usuarios["usuarios"].get(nick_normalizado,None)

def get_openid(conn,usuario,OpenID) :
  OpenIDs=conn.get_root()["usuarios"]["usuarios"].get(normaliza_nick(usuario))
  if not OpenIDs : return None
  OpenIDs=OpenIDs["OpenID"]
  if OpenID in OpenIDs : return True
  if OpenID[-1]=="/" :
    return OpenID[:-1] in OpenIDs
  else :
    return OpenID+"/" in OpenIDs

def grupo_add(conn,nombre) :
  from durus.persistent_dict import PersistentDict

  root=conn.get_root()
  grupos=root["grupos"]
  num_groups=grupos["num_grupos"]
  grupos["num_grupos"]+=1
  mt=metatag_add(conn,"Mensajes privados del grupo "+nombre,private=True)
  g=PersistentDict({"nombre":nombre,
                    "usuarios":set(),
                    "metatag_privado":mt,
                  })
  grupos["grupos"][num_groups]=g
  return num_groups

def metatag_add(conn,nombre,usuarios_moderadores=None,grupos_moderadores=None,private=False) :
  from durus.btree import BTree
  from durus.persistent_dict import PersistentDict

  if usuarios_moderadores==None : usuarios_moderadores=[]
  if grupos_moderadores==None : grupos_moderadores=set()

  root=conn.get_root()
  metatags=root["metatags"]
  num_mt=metatags["num_metatags"]
  metatags["num_metatags"]+=1
  d=PersistentDict({"nombre":nombre,
                    "num_hilos":0,
                    "TS2hilo":BTree(),
                    "usuarios moderadores":usuarios_moderadores,
                    "grupos moderadores":grupos_moderadores,
                    "monitor":BTree(),
                    "sticky_threads":set(),
                   })

  metatags["metatags"][num_mt]=d

  if not private :
    metatags["publicos"].add(num_mt)
    metatags["categorias"].append([num_mt])
    for usuario in root["usuarios"]["usuarios"].itervalues() :
      usuario["punto de lectura no leidos"][num_mt]=BTree()

  return num_mt

def metatag_add_user(conn,metatag,usuario) :
  root=conn.get_root()
  usuario=root["usuarios"]["usuarios"][normaliza_nick(usuario)]
  plnl=usuario["punto de lectura no leidos"]
  if metatag not in plnl :
    from durus.btree import BTree
    plnl[metatag]=BTree()

def metatag_del_user(conn,metatag,usuario) :
  root=conn.get_root()
  usuario=root["usuarios"]["usuarios"][normaliza_nick(usuario)]
  assert usuario["metatag_privado"]!=metatag
  plnl=usuario["punto de lectura no leidos"]
  if metatag in plnl :
    del plnl[metatag]

def grupo_add_user(conn,grupo,usuario) :
  root=conn.get_root()
  metatag=root["grupos"]["grupos"][grupo]["metatag_privado"]
  metatag_add_user(conn,metatag,usuario)
  u=root["usuarios"]["usuarios"][normaliza_nick(usuario)]
  u["grupos"].add(grupo)
  u._p_note_change()  # MUTABLE

def grupo_del_user(conn,grupo,usuario) :
  root=conn.get_root()
  metatag=root["grupos"]["grupos"][grupo]["metatag_privado"]
  metatag_del_user(conn,metatag,usuario)
  u=root["usuarios"]["usuarios"][normaliza_nick(usuario)]
  u["grupos"].discard(grupo)
  u._p_note_change()  # MUTABLE

def get_message_source(conn,num_msg) :
  root=conn.get_root()
  msg=root["mensajes"]["mensajes"].get(num_msg)
  if not msg : return None,None
  return (msg["autor"],msg["texto"][0])

def mensaje_add(conn,texto_bruto,texto_procesado,nick,hilo=None,titulo=None,metatags=None) :
  from durus.btree import BTree
  from durus.persistent_dict import PersistentDict
  import time

  usuario=normaliza_nick(nick)

  root=conn.get_root()
  hilos=root["hilos"]
  mensajes=root["mensajes"]
  seguimiento_no_leidos=root["seguimiento_no_leidos"]
  metatags2=root["metatags"]["metatags"]

  ts=time.time()
  mensaje=PersistentDict(
             {"texto":[texto_bruto,texto_procesado],
              "TS":ts,
              "autor":nick})

  mensajes["num_mensajes"]+=1
  num_msg=mensajes["num_mensajes"]
  mensajes["mensajes"][num_msg]=mensaje

  if hilo!=None :
    valor_retorno=(hilo,num_msg)
    h=hilo
    mensaje["hilo"]=hilo
    hilo=hilos["hilos"][hilo]
    mon=[hilo["monitor"]]
    last=hilo["mensajes"].get_max_item()[0]
    del seguimiento_no_leidos["last_msg2hilo"][last]
    TS=hilo["TS"]
    for metatag in hilo["metatags"] :
      del metatags2[metatag]["TS2hilo"][TS]
  else : # Empezamos un hilo nuevo
    from durus.persistent_set import PersistentSet
    hilos["num_hilos"]+=1
    h=hilos["num_hilos"]
    valor_retorno=(h,num_msg)
    root["punto_de_lectura_global"][h]=0
    mon=[]
    for metatag in metatags :
      mon.append(root["metatags"]["metatags"][metatag]["monitor"])
    hilo=PersistentDict()
    assert titulo
    hilo=PersistentDict(
         {"creador":nick,
          "TS creacion":ts,
          "titulo":titulo,
          "metatags":PersistentSet(metatags),
          "monitor":BTree(),
          "num_mensajes":0,
          "mensajes":BTree(),
          "numeros":BTree()})
    hilos["hilos"][h]=hilo
    mensaje["hilo"]=h
    last=0  # Simplify mailing

  seguimiento_no_leidos["hilo2last_msg"][h]=num_msg
  seguimiento_no_leidos["last_msg2hilo"][num_msg]=h

  hilo["num_mensajes"]+=1
  num_mensajes=hilo["num_mensajes"]
  hilo["mensajes"][num_msg]=(num_mensajes,mensaje)
  hilo["numeros"][num_mensajes]=num_msg
  hilo["TS"]=ts
  for metatag in hilo["metatags"] :
    metatags2[metatag]["TS2hilo"][ts]=(h,hilo)

  usuario=root["usuarios"]["usuarios"][usuario]
  usuario["num mensajes"]+=1
  usuario["mensajes"][num_msg]=mensaje

  meta_u=root["usuarios"]["usuarios"]
  for mon in mon :
    destinations=[]
    for i in mon.iterkeys() :
      u=meta_u[normaliza_nick(i)]
      umc=u["ultimo mensaje conocido"]
      if umc<last : continue  # Already notified
      plnl=u["punto de lectura no leidos"]
      already_notified=False
      for metatag in hilo["metatags"] :
        if metatag not in plnl : continue   # No tenemos acceso a ese metatag
        if h in plnl[metatag] :
          already_notified=True
          break  # Already notified
      if already_notified : continue
      email=u["email"]
      if email : destinations.append(email)
    if len(destinations) :
      from email.header import Header
      import smtp
      texto=["From: %s" %globales.smtp_email_sender]
      texto.append("To: %s" %globales.smtp_email_sender)

      if last : # Thread
        texto.append("Subject: Nuevo mensaje en el hilo %s"
          %Header("'%s'" %hilo["titulo"],"utf-8",76,"Subject: Nuevo mensaje en el hilo ").encode())
        texto.append("")
        texto.append("Nuevo mensaje")
        texto.append("%s/hilo/%d/%d#m%d" %(globales.base_url,h,1+globales.thread_len*((num_mensajes-1)/globales.thread_len),num_msg))
      else : # Metatag
        texto.append("Subject: Nuevo hilo %s"
          %Header("'%s'" %hilo["titulo"],"utf-8",76,"Subject: Nuevo hilo ").encode())
        texto.append("")
        texto.append("Nuevo hilo")
        texto.append("%s/hilo/%d" %(globales.base_url,h))

      smtp.enqueue(conn,destinations,"\r\n".join(texto))

  return valor_retorno

def actualiza_no_leidos(conn,usuario) :
  root=conn.get_root()
  global_last=root["mensajes"]["num_mensajes"]
  if not usuario : # Acceso anonimo. Nos encargamos de que nada salga como "no leido"
    usuario=root["usuarios"]["usuarios"][usuario]
    if usuario["ultimo mensaje conocido"]!=global_last :
      usuario["ultimo mensaje conocido"]=global_last
    return

  usuario=root["usuarios"]["usuarios"][normaliza_nick(usuario)]
  l=usuario["ultimo mensaje conocido"]
  if l==global_last : return   # No hay nada nuevo

  punto_de_lectura_global=root["punto_de_lectura_global"]
  posicion_hilos_no_leidos=dict(usuario["punto de lectura no leidos"])
  posicion_hilos=usuario["punto de lectura"]

  last_msg2hilo=root["seguimiento_no_leidos"]["last_msg2hilo"]
  hilos=root["hilos"]["hilos"]

  for last,hilo in last_msg2hilo.items_from(l,closed=False) :
    # Vemos cual es el ultimo leido para ese hilo
    ultimo=posicion_hilos.get(hilo) # Esta es la optimizacion de hilos que todo el mundo ha leido ya...
    if not ultimo : ultimo=punto_de_lectura_global[hilo]
    if ultimo < last : # Este hilo tiene mensajes nuevos...
      # Lo marcamos como pendiente de leer y nos vamos al primero no leido
      h=hilos[hilo]
      ultimo=h["mensajes"].items_from(ultimo,closed=False).next()[0]
      for metatag in h["metatags"] :
        mt=posicion_hilos_no_leidos.get(metatag)
        if mt!=None : mt[hilo]=ultimo
  usuario["ultimo mensaje conocido"]=global_last

def listado_stickies_personal(conn,usuario,metatag) :
  root=conn.get_root()
  hilos=root["hilos"]["hilos"]
  usuario=root["usuarios"]["usuarios"][normaliza_nick(usuario)]
  hilo2last_msg=root["seguimiento_no_leidos"]["hilo2last_msg"]

  no_leidos=usuario["punto de lectura no leidos"][metatag]
  monitor_hilos=usuario["monitor hilos"]

  resultado=[]

  metatag=root["metatags"]["metatags"][metatag]
  stickies=metatag["sticky_threads"]
  r=[(hilo2last_msg[hilo_num],hilo_num) for hilo_num in stickies]
  r.sort()
  r.reverse()
  for dummy,hilo_num in r :
    hilo=hilos[hilo_num]
# If we modified the data returned, we must adjust also the peer "plain threads" function.
    resultado.append((hilo_num,hilo["titulo"],hilo_num in no_leidos,hilo_num in monitor_hilos,"read only" in hilo))

  return resultado

def listado_hilos_personal(conn,usuario,ts,ts2,metatag) :
  root=conn.get_root()
  hilos=root["hilos"]["hilos"]
  usuario=root["usuarios"]["usuarios"][normaliza_nick(usuario)]

  no_leidos=usuario["punto de lectura no leidos"][metatag]
  monitor_hilos=usuario["monitor hilos"]

  resultado=[]

  metatag=root["metatags"]["metatags"][metatag]
  stickies=metatag["sticky_threads"]
  for ts,hilo in metatag["TS2hilo"].items_range(ts,ts2) :
    hilo_num,hilo=hilo
    if hilo_num in stickies : continue
# If we modified the data returned, we must adjust also the peer "sticky" function.
    resultado.append((hilo_num,hilo["titulo"],hilo_num in no_leidos,hilo_num in monitor_hilos,"read only" in hilo))

  return resultado

def listado_hilo(conn,usuario,hilo_num,primer_mensaje=1) :
  root=conn.get_root()
  hilo=root["hilos"]["hilos"].get(hilo_num,None)
  if not hilo : return None,None,None,None,None,None

  meta={"titulo":hilo["titulo"],"creador":hilo["creador"],
        "TS creacion":hilo["TS creacion"],"read only":("read only" in hilo)}

  avatares=root["static"]["avatars"]
  usuarios=root["usuarios"]["usuarios"]
  u=usuarios[normaliza_nick(usuario)]
  mon=hilo_num in u["monitor hilos"]

  num_mensajes=hilo["num_mensajes"]
  mensajes=[]
  i=globales.thread_len
  pr=hilo["numeros"].get(primer_mensaje,None)
  if not pr : return None,None,None,None,None,None
  iterador_mensajes=hilo["mensajes"].items_from(pr)
  for msg_num,msg in iterador_mensajes :
    posicion,msg=msg
    autor=msg["autor"]
    autor2=usuarios.get(normaliza_nick(autor))
    avatar=None
    if autor2 :
      avatar=autor2["avatar"]
      if avatar :
        avatar=(avatar,avatares[avatar][0]["size"])
    mensajes.append((msg_num,autor,msg["TS"],msg["texto"][1],avatar))
    i-=1
    if not i : break

  metatags=hilo["metatags"]
  finished=posicion>=num_mensajes
  if usuario : # Acceso *NO* anonimo
    pl=u["punto de lectura"].get(hilo_num,0)
    if not pl : # Nunca hemos visitado ese hilo, o hace mucho que no cambia
      pl=root["punto_de_lectura_global"][hilo_num]
    if pl<msg_num :
      u["punto de lectura"][hilo_num]=msg_num
      pnl=0
      for metatag in metatags :
        plnl=u["punto de lectura no leidos"].get(metatag)
        if not plnl : continue  # No tenemos acceso a ese metatag
        nl=plnl.get(hilo_num)
        if nl and (nl<=msg_num) :
          if finished : del plnl[hilo_num]
          else :
            if not pnl :
              pnl=iterador_mensajes.next()[0] # El primer no leido
            plnl[hilo_num]=pnl

  mt=root["metatags"]["metatags"]
  return (meta,mensajes,num_mensajes,finished,[(i,mt[i]["nombre"]) for i in metatags],mon)

# Marcamos como leidos los hilos que estamos
# viendo en pantalla. Si hay mensajes adicionales
# que todavia no hemos visualizado en el indice,
# los mantenemos como "no leidos".
def marca_todo_leido(conn,usuario,metatag) :
  usuario=conn.get_root()["usuarios"]["usuarios"][normaliza_nick(usuario)]
  posicion_hilos_no_leidos=usuario["punto de lectura no leidos"][metatag]
  posicion_hilos=usuario["punto de lectura"]
 
  for hilo,last_msg in posicion_hilos_no_leidos.items() :
    posicion_hilos[hilo]=last_msg

  posicion_hilos_no_leidos.clear()

def user_has_metatag_access(conn,usuario,metatag,write=False) :
  usuario=conn.get_root()["usuarios"]["usuarios"][normaliza_nick(usuario)]
  if write and (metatag==usuario["metatag_privado"]) : return False  # Los mensajes privados no se escriben directamente en el metatag
  return metatag in usuario["punto de lectura no leidos"]

def user_has_thread_access(conn,usuario,thread,write=False) :
  root=conn.get_root()
  u=root["usuarios"]["usuarios"][normaliza_nick(usuario)]
  plnl=u["punto de lectura no leidos"]
  thread=root["hilos"]["hilos"].get(thread)
  if thread==None : return False
  for metatag in thread["metatags"] :
    if metatag in plnl :
      if metatag==u["metatag_privado"] :
        return globales.allow_private_messages
      return True
  return False

def get_metatags(conn,usuario) :
  valor=[]
  root=conn.get_root()
  metatags=root["metatags"]
  categorias=metatags["categorias"]
  metatags=metatags["metatags"]
  u=root["usuarios"]["usuarios"][normaliza_nick(usuario)]
  plnl=u["punto de lectura no leidos"]
  u_metatags=plnl.keys()
  monitor_metatags=u["monitor metatags"]

  for i in categorias :
    if len(i)==1 :
      i=i[0]
      plnl2=plnl.get(i)
      if plnl2!=None :
        u_metatags.remove(i)
        valor.append([(i,metatags[i]["nombre"], True if plnl2 else False, i in monitor_metatags)])
    else :
      cat=[i[0]]
      for j in i[1:] :
        plnl2=plnl.get(j)
        if plnl2!=None :
          u_metatags.remove(j)
          cat.append((j,metatags[j]["nombre"], True if plnl2 else False, j in monitor_metatags))
      if len(cat)>1 :
        valor.append(cat)

  cat=["Area privada"]
  i=u["metatag_privado"]
  if i in u_metatags :
    u_metatags.remove(i)
    plnl2=plnl[i]
    cat.append((i,metatags[i]["nombre"], True if plnl2 else False, i in monitor_metatags))
  for i in u_metatags :
    u_metatags.remove(i)
    plnl2=plnl[i]
    cat.append((i,metatags[i]["nombre"], True if plnl2 else False, i in monitor_metatags))
  if len(cat)>1 :
    valor.append(cat)

  assert len(u_metatags)==0

  return valor

def monitored_metatag(conn,usuario,metatag) :
  return metatag in conn.get_root()["usuarios"]["usuarios"][normaliza_nick(usuario)]["monitor metatags"]

def toggle_monitor_metatag(conn,usuario,metatag) :
  root=conn.get_root()
  nick_norm=normaliza_nick(usuario)
  u=root["usuarios"]["usuarios"][nick_norm]

  if metatag in u["monitor metatags"] :
    if metatag==u["metatag_privado"] : return None,True
    del u["monitor metatags"][metatag]
    del root["metatags"]["metatags"][metatag]["monitor"][nick_norm]
    return None,False

  if metatag not in u["punto de lectura no leidos"] : # Check permission
    return "Too clever for your own good...",False

  root["metatags"]["metatags"][metatag]["monitor"][nick_norm]=None
  u["monitor metatags"][metatag]=None
  return None,True

def toggle_monitor_thread(conn,usuario,thread) :
  root=conn.get_root()
  nick_norm=normaliza_nick(usuario)
  u=root["usuarios"]["usuarios"][nick_norm]

  if thread in u["monitor hilos"] :
    thr=root["hilos"]["hilos"][thread]
    if u["metatag_privado"] in thr["metatags"] :
      return None,True

    del u["monitor hilos"][thread]
    del thr["monitor"][nick_norm]
    return None,False

  thr=root["hilos"]["hilos"].get(thread)
  if not thr : return "Too clever for your own good...",False
  plnl=u["punto de lectura no leidos"]
  for i in thr["metatags"] :
    if i in plnl : # Check permission
      break
  else :
    return "Too clever for your own good...",False

  thr["monitor"][usuario]=None
  u["monitor hilos"][thread]=None
  return None,True

def check_thread_moderator_or_master_status(conn,usuario,num_thread) :
  root=conn.get_root()
  hilos=root["hilos"]["hilos"]
  metatags=root["metatags"]["metatags"]
  if num_thread not in hilos : return False  # Thread doesn't exist
  usuario=normaliza_nick(usuario)
  if usuario=="master" : return True
  grupos=root["usuarios"]["usuarios"][usuario]["grupos"]
  for metatag in hilos[num_thread]["metatags"] :
    for moderador in metatags[metatag]["usuarios moderadores"] :
      if usuario==normaliza_nick(moderador) : return True
    for grupo in metatags[metatag]["grupos moderadores"] :
      if grupo in grupos : return True
  return False

def check_permissions_readonly(conn,usuario,num_thread) :
  return check_thread_moderator_or_master_status(conn,usuario,num_thread)

def check_permissions_sticky(conn,usuario,num_thread) :
  return check_thread_moderator_or_master_status(conn,usuario,num_thread)

def thread_is_readonly(conn,hilo) :
  root=conn.get_root()
  hilos=root["hilos"]["hilos"]
  return (hilo not in hilos) or ("read only" in hilos[hilo])

def toogle_thread_readonly(conn,num_thread) :
  root=conn.get_root()
  hilo=root["hilos"]["hilos"][num_thread]
  if "read only" in hilo :
    del hilo["read only"]
    return False
  hilo["read only"]=None
  return True


def toogle_thread_sticky(conn,num_thread) :
  root=conn.get_root()
  hilo=root["hilos"]["hilos"][num_thread]
  metatags=root["metatags"]["metatags"]
  somebody_sticky=False
  somebody_no_sticky=False
  for metatag in hilo["metatags"] :
    metatag=metatags[metatag]
    sticky=metatag["sticky_threads"]
    if num_thread in sticky :
      somebody_sticky=True
      sticky.discard(num_thread)
    else :
      somebody_no_sticky=True
      sticky.add(num_thread)
    metatag._p_note_change() # MUTABLE

  assert somebody_sticky!=somebody_no_sticky
  return not somebody_sticky

