# $Id: parsers.py 497 2007-11-13 21:50:56Z alvaro $

__doc__ == """
 The parsers must return a tuple in the form:
 (position, string, tag, arguments, string_after)
 where:
   - Position: the position of the tag in the text.
   - String: String to be added to the stack (to be
       treated as TEXT.
   - Tag: The parsed tag. To be treated as a TAG.
       Has the format: (TYPE, True/Flase, Tag, nargs),
       where TYPE is the type of the tag (BBCODE, HTML,
       etc.), Is Opening is True when the tag is an opening
       one. The last item is the tag name. Nargs is the number
       of arguments that the tag needs,
   - Arguments: A list with the arguments for the tag.
   - String After: The string after the tag (i.e. the
       remainder of the text.

 The master parser must call each parser individually
 and choose the one which returns the leftmost tag. Then
 redo this operation with the remainder text.
"""

import parser_html
import parser_urls
import parser_bbcode
import parser_smileys
import parser_eol

def gimme_the_tag(tag,args,nargs=0):
    mapping = {
            "a": ('<a href="%s">%s</a>'),
            "url": ('<a href="%s">%s</a>'),
            "b": ('<strong>%s</strong>'),
            "u": ('<span style="text-decoration:underline">%s</span>'),
            "s": ('<strile>%s</strike>'),
            "size": ('<span style="font-size:%s">%s</span>'), 
            "code": ('<code>%s</code>'), 
            "i": ('<em>%s</em>'), 
            "quote*": ('<blockquote>%s</blockquote>'),
            "quote": ('<blockquote title="%s"><h4>%s escribi&oacute;:</h4>%s</blockquote>'),
            "img": ('<img src="%s" />')
        }
        
    aux = args[0:nargs]
    aux.append("".join(args[nargs:]))
    args = aux
    if tag == "url" and len(args) < 2:
        args.append(args[-1])
    elif tag == "quote" and len(args) == 1:
        tag  = "quote*"
    elif tag == "quote" and len(args) == 2:
        args.insert(0,args[0])
    return mapping[tag] % tuple(args)

def escape(text):
  import cgi
  return cgi.escape(text)

def convert_to_html(text):
    """Converts the text into valid HTML."""

    import globales
    parsers = [
            (True, parser_urls),
            (globales.allow_html, parser_html),
            (globales.allow_bbcode, parser_bbcode),
            (True, parser_smileys), # FIXME
            (True, parser_eol),
    ]
    
    
    stack = []
    context = [None,]

    while True:
        results = {}

        for allow, parser in parsers:
            ret =  parser.parse(text,context[-1])
            results[ret[0]] = ret[1:]

        # Choose the parser with the leftmost coincidence
        if min(results.keys()) == -1:
            return True, results[min(results.keys())][0]
        before, tag, args, after = results[min(results.keys())]
        
        # Append the text _before_ the tag to the stack, type TEXT
        stack.append(('TEXT',escape(before)))
        
        if tag and not tag[1]: # It is a closing
            args = []
            while True: # Pop items until we find the opening
                try:
                    el = stack.pop()
                except: # We have no tag to close
#                    print tag
#                    print stack
                    return True, 'Texto mal formateado'
                if el[0] == 'TEXT': # Threat this as the arguments for the tag until we find it.
                    args.insert(0,el[1])
                elif el[0] == tag[0] and el[2] == tag[2]: # I have the tag that matches
                    stack.append(('TEXT', gimme_the_tag(el[2], args, el[3])))
                    context.pop()
                    break
                else: # We have no tag to close
#                    print tag
#                    print stack
                    return True, 'Texto mal formateado'
        elif tag: # It is an opening
            stack.append(tag)
            if tag[0] == "TEXT":
                pass
            else:
                context.append(tag[2])
            if args:
                args = [escape(i) for i in args]
#                args = escape(args)
                for i in args:
                  stack.append(('TEXT',i))
        if after:
            text = after
        else:
            break
    
    text = []
    while True:
        try:
            el = stack.pop(0)
            if el[0] == 'TEXT':
                text.append(el[1])
            else:
#                print el
#                print stack
                return True, "Texto mal formateado"
        except:
            return False, "".join(text)


import unittest
class TestBBCodeHTML(unittest.TestCase):
    """Test Case for the BBCode and HTML parser together"""
    correct_bbcode = (

            ('[ b]hola[/b]', '<strong>hola</strong>'), 
            ('[b ]hola[/b]', '<strong>hola</strong>'), 
            ('[ b ]hola[/b]', '<strong>hola</strong>'), 

            ('[b]hola[ /b]', '<strong>hola</strong>'), 
            ('[b]hola[/b ]', '<strong>hola</strong>'), 
            ('[b]hola[ /b ]', '<strong>hola</strong>'), 

            ('[b]hola[/b]', '<strong>hola</strong>'), 
            ('antes [b]hola[/b]', 'antes <strong>hola</strong>'), 
            ('[b]hola[/b] despues', '<strong>hola</strong> despues'), 
            ('antes [b]hola[/b] despues', 'antes <strong>hola</strong> despues'), 
            
            ('[url]la direccion[/url]', '<a href="la direccion">la direccion</a>'), 
            ('antes [url]la direccion[/url]', 'antes <a href="la direccion">la direccion</a>'), 
            ('[url]la direccion[/url] despues', '<a href="la direccion">la direccion</a> despues'), 
            ('antes [url]la direccion[/url] despues', 'antes <a href="la direccion">la direccion</a> despues'), 
            
            ('[url="la direccion"]el texto[/url]', '<a href="la direccion">el texto</a>'), 
            ('antes [url="la direccion"]el texto[/url]', 'antes <a href="la direccion">el texto</a>'), 
            ('[url="la direccion"]el texto[/url] despues', '<a href="la direccion">el texto</a> despues'), 
            ('antes [url="la direccion"]el texto[/url] despues', 'antes <a href="la direccion">el texto</a> despues'), 

            ('[url="la direccion"]el texto[/url]', '<a href="la direccion">el texto</a>'), 
            
            ('[url="http://example.org/uno&dos"]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            
            ('[url ="http://example.org/uno&dos"]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[url= "http://example.org/uno&dos"]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[url = "http://example.org/uno&dos"]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[ url="http://example.org/uno&dos"]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[ url ="http://example.org/uno&dos"]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[ url = "http://example.org/uno&dos"]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[ url= "http://example.org/uno&dos"]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[url="http://example.org/uno&dos" ]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[ url="http://example.org/uno&dos" ]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[ url = "http://example.org/uno&dos" ]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            
            ('[ url = http://example.org/uno&dos]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[url=http://example.org/uno&dos]el texto[ /url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[url=http://example.org/uno&dos]el texto[ / url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            ('[url=http://example.org/uno&dos]el texto[ /url ]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            
            ('[url=la direccion]el texto[/url]', '<a href="la direccion">el texto</a>'), 
            ('antes [url=la direccion]el texto[/url]', 'antes <a href="la direccion">el texto</a>'), 
            ('[url=la direccion]el texto[/url] despues', '<a href="la direccion">el texto</a> despues'), 
            ('antes [url=la direccion]el texto[/url] despues', 'antes <a href="la direccion">el texto</a> despues'), 
            
            ('[quote=paco]Hola[/quote]', '<blockquote title="paco"><h4>paco escribi&oacute;:</h4>Hola</blockquote>'), 
            ('antes [quote=paco]Hola[/quote]', 'antes <blockquote title="paco"><h4>paco escribi&oacute;:</h4>Hola</blockquote>'), 
            ('[quote=paco]Hola[/quote] despues', '<blockquote title="paco"><h4>paco escribi&oacute;:</h4>Hola</blockquote> despues'), 
            ('antes [quote=paco]Hola[/quote] despues', 'antes <blockquote title="paco"><h4>paco escribi&oacute;:</h4>Hola</blockquote> despues'), 
            
            ('[quote]Hola[/quote]', '<blockquote>Hola</blockquote>'), 
            ('antes [quote]Hola[/quote]', 'antes <blockquote>Hola</blockquote>'), 
            ('[quote]Hola[/quote] despues', '<blockquote>Hola</blockquote> despues'), 
            ('antes [quote]Hola[/quote] despues', 'antes <blockquote>Hola</blockquote> despues'),            
            
            ('[u]hola[/u]', '<span style="text-decoration:underline">hola</span>'), 
      
            ('[size=20px]hola[/size]', '<span style="font-size:20px">hola</span>'), 

            ('[n]hola[/n]', '[n]hola[/n]'), 

            ("[img]laimagen1[/img]", '<img src="laimagen1" />'), 
            ("antes [img]laimagen2[/img]", 'antes <img src="laimagen2" />'), 
            ("[img]laimagen3[/img] despues", '<img src="laimagen3" /> despues'), 
            ("antes [img]laimagen4[/img] despues", 'antes <img src="laimagen4" /> despues'),

            ('[url="http://example.org/uno&dos"]el texto[/url]', '<a href="http://example.org/uno&amp;dos">el texto</a>'), 
            
            
            ('[url="http://example.org/uno&dos"]el & texto[/url]', '<a href="http://example.org/uno&amp;dos">el &amp; texto</a>'), 
            ('[url="http://example.org/uno&dos"]el &amp; texto[/url]', '<a href="http://example.org/uno&amp;dos">el &amp;amp; texto</a>'), 
            ('[url="http://example.org/uno&dos"]el "" texto[/url]', '<a href="http://example.org/uno&amp;dos">el "" texto</a>'), 
            ('[url="http://example.org/uno&dos"]el "" texto[/url]', '<a href="http://example.org/uno&amp;dos">el "" texto</a>'), 

            ('[url="http://example.org/uno&dos"]el texto[b]en negrita[/b][/url]', '<a href="http://example.org/uno&amp;dos">el texto<strong>en negrita</strong></a>'), 
            ('[quote]Hola[quote]otro[/quote]Adios[/quote]', '<blockquote>Hola<blockquote>otro</blockquote>Adios</blockquote>'), 
            
            ('[quote]Hola[url]http://example.com[/url]Adios[/quote]', '<blockquote>Hola<a href="http://example.com">http://example.com</a>Adios</blockquote>'), 

            ('[quote="paco"]Hola[quote]otro[/quote]Adios[/quote]', '<blockquote title="paco"><h4>paco escribi&oacute;:</h4>Hola<blockquote>otro</blockquote>Adios</blockquote>'), 
    )
    correct_html = (
            ('<b>foo</b>', '<strong>foo</strong>'), 
            ('b>', 'b&gt;'), 
            ('<script>', '&lt;script&gt;'), 
            ('<a href="http://blablabla.com/index.php?foo=1&bar=2">bla bla</a>', 
             '<a href="http://blablabla.com/index.php?foo=1&amp;bar=2">bla bla</a>'), 
            ('<a href="http://blablabla.com/index.php?foo=1&bar=2" title="foo">bla bla </a>', 
             '<a href="http://blablabla.com/index.php?foo=1&amp;bar=2">bla bla </a>'), 
            ('<img src="http://blablabla.com/img.php?foo=1&bar=2" />', 
             '<img src="http://blablabla.com/img.php?foo=1&amp;bar=2" />'), 
            ('<img src="http://blablabla.com/img.php?foo=1&bar=2" alt="foo">', 
             '<img src="http://blablabla.com/img.php?foo=1&amp;bar=2" alt="foo" />'), 
            ('<img src="http://blablabla.com/img.php?foo=1&bar=2" alt="foo"></img>', 
             '<img src="http://blablabla.com/img.php?foo=1&amp;bar=2" alt="foo" />'),
            ('&amp;', '&amp;'), 
    )
    correct_urls = (
            ('http://example.com', ' <a href="http://example.com" title="http://example.com">http://example.com</a> '),
            ('antes http://example.com', 'antes <a href="http://example.com" title="http://example.com">http://example.com</a> '),
            ('http://example.com despues', ' <a href="http://example.com" title="http://example.com">http://example.com</a> despues'),
            ('antes http://example.com despues', 'antes <a href="http://example.com" title="http://example.com">http://example.com</a> despues'),
    )
    correct = (
            ('http://example.com [url="http://example.org/uno&dos"]el "" texto[/url]', ' <a href="http://example.com" title="http://example.com">http://example.com</a> <a href="http://example.org/uno&amp;dos">el "" texto</a>'), 
    )
    incorrect = (
            '<a href="caca>paco<a>luis</a>manolo</a>ringesvinto',
            '<a>paco<a>luis</a>manolo</a>ringesvinto',
            '<a>paco[url]luis[/url]>manolo</a>ringesvinto',
            '[url]paco<a>luis</a>manolo[/url]ringesvinto',
            '<b>', 
            '</b>', 
            '<a href="http://blablabla.com/index.php?foo=1&bar=2">bla bla', 
            '<a href="http://blablabla.com/index.php?foo=1&bar=2">bla bla </as>', 
            '<a href="http://blablabla.com/index.php?foo=1&bar=2" title="foo" de la vaca>bla bla <a>', 
            '<img sorcerer="http://blablabla.com/index.php?foo=1&bar=2" />', 

    )

    def testIncorrectCode(self):
        """Test if the translation goes well."""
        global allow_errors
        allow_errors = False
        for code in self.incorrect:
            retval, dummy = convert_to_html(code)
            assert retval

    def testCorrectURLSCode(self):
        """Test if the translation goes well."""
        global allow_errors
        allow_errors = False
        for code, html in self.correct_urls:
            retval, result = convert_to_html(code)
            assert not retval
            self.assertEqual(html, result)

    def testCorrectBBCode(self):
        """Test if the translation goes well."""
        global allow_errors
        allow_errors = False
        for code, html in self.correct_bbcode:
            retval, result = convert_to_html(code)
            assert not retval
            self.assertEqual(html, result)

    def testCorrectHTML(self):
        """Test if the translation goes well."""
        global allow_errors
        allow_errors = False
        for code, html in self.correct_html:
            retval, result = convert_to_html(code)
            assert not retval
            self.assertEqual(html, result)

    def testCorrectCode(self):
        """Test if the translation goes well."""
        global allow_errors
        allow_errors = False
        for code, html in self.correct:
            retval, result = convert_to_html(code)
            assert not retval
            self.assertEqual(html, result)
  
if __name__ == "__main__":
    import unittest
    unittest.main()
